@@ -9,11 +9,11 @@ Geo::Google - Perform geographical queries using Google Maps
   use Geo::Google;
 
   #Allen's office
-  my $gonda_addr = '695 Charles E Young Dr S, Westwood, CA 90024';
+  my $gonda_addr = '695 Charles E Young Dr S, Los Angeles, Los Angeles, California 90024, United States';
   #Stan's Donuts
   my $stans_addr = '10948 Weyburn Ave, Westwood, CA 90024';
   #Roscoe's House of Chicken and Waffles
-  my $roscoes_addr = "5006 W Pico Blvd, Los Angeles, CA";
+  my $roscoes_addr = "5006 W Pico Blvd, Los Angeles, CA 90019";
 
   #Instantiate a new Geo::Google object.
   my $geo = Geo::Google->new();
@@ -118,8 +118,6 @@ send a patch.  Known bugs/issues:
 
 =over
 
-=item Polyline decoding needs to be cleaned up. 
-
 =item Lack of documentation.
 
 JSON exporting is not exactly identical to the original Google 
@@ -140,7 +138,7 @@ interchangeable with a Google JSON reponse.
 
 package Geo::Google;
 use strict;
-our $VERSION = '0.03';
+our $VERSION = '0.04';
 
 #this gets a javascript page containing map XML
 use constant LQ => 'http://maps.google.com/maps?output=js&v=1&q=%s';
@@ -234,47 +232,57 @@ sub error {
 
 sub location {
   my ( $self, %arg ) = @_;
+  my @result = ();
 
   my $address   = $arg{'address'} or ($self->error("must provide an address to location()") and return undef);
 
-  my $json = new JSON (skipinvalid => 1, barekey => 1, 
-			quotapos => 1, unmapping => 1 );
+  my $json = new JSON (skipinvalid => 1, barekey => 1, quotapos => 1, unmapping => 1 );
   my $response_json = undef;
-  my $page = get( sprintf( LQ, uri_escape($address) ) );
-
+  # I'm using an an array here because I might need to parse several pages if Google suggests a different address
+  my @pages = ( get( sprintf( LQ, uri_escape($address) ) ) );
+  
   # See if google returned no results
-  if ( $page =~ /did\snot\smatch\sany\slocations/i ) {
-    $self->error( "Google couldn't match any locations matching "
-	. "$address.");
-    return undef;
+  if ( $pages[0] =~ /did\snot\smatch\sany\slocations/i ) {
+    $self->error( "Google couldn't find any locations matching $address." ) and return undef;
   }
-  # attept to locate the JSON formatted data block
-  elsif ($page =~ 	
-	/loadVPage\((.+),document\.getElementById\(/is) {
-	my $strJSON = $1;
-	$response_json = $json->jsonToObj($strJSON);
-	}
-  else {
-	$self->error( "Unable to locate the JSON format data in " 
-		. "google's response.");
-	return undef;
-	}
-
-  if ( scalar(@{$response_json->{"overlays"}->{"markers"}}) > 0 ) {
-	my @result = ();
-	foreach my $marker (@{$response_json->{"overlays"}->{"markers"}}) {
-		my $loc = $self->_obj2location($marker, %arg);
-		push @result, $loc;
-	}		
-  	return @result;
+  # See if Google was unable to resolve the address, but suggested other addresses
+  # To see this, run a query for 695 Charles E Young Dr S, Westwood, CA 90024
+  elsif ( $pages[0] =~ m#Did you mean:#is ) {
+    # Extract the queries from all the http get queries for alterate addresses
+    # \u003cdiv class=\"ref\"\u003e\u003ca href=\"/maps?v=1\u0026amp;q=695+Charles+E+Young+Drive+East,+Los+Angeles,+Los+Angeles,+California+90024,+United+States\u0026amp;ie=UTF8\u0026amp;hl=en\u0026amp;oi=georefine\u0026amp;ct=clnk\u0026amp;cd=2\" onclick=\"return loadUrl(this.href)\"\u003e
+    # We need it to fit the LQ query 'http://maps.google.com/maps?output=js&v=1&q=%s'
+    my @queries = $pages[0] =~ m#\\u003cdiv class=\\"ref\\"\\u003e\\u003ca href=\\"/maps\?v=1\\u0026amp;q=(.+?)\\u0026amp;#gsi;
+    # clear the $pages array so we can fill it with the pages from the @urls
+    @pages = ();
+    foreach my $suggested_query (@queries) {
+      push( @pages, get( sprintf( LQ, $suggested_query ) ) );
+    }
+  }
+  # Verify that we actually retrieved pages to parse
+  if ( scalar(@pages) > 0 ) {
+    foreach my $page (@pages) {
+      # attempt to locate the JSON formatted data block
+      if ($page =~ m#loadVPage\((.+), "\w+"\);}//]]>#is) { $response_json = $json->jsonToObj($1); }
+      else {
+	$self->error( "Unable to locate the JSON format data in google's response.") and return undef;
+      }
+      if ( scalar(@{$response_json->{"overlays"}->{"markers"}}) > 0 ) {	
+        foreach my $marker (@{$response_json->{"overlays"}->{"markers"}}) {
+	  my $loc = $self->_obj2location($marker, %arg);
+	  push @result, $loc;
+        }		
+      }
+      else {
+	$self->error("Found the JSON Data block and was able to parse it, but it had no location markers "
+	  . "in it.  Maybe Google changed their JSON data structure?.") and return undef;
+      }
+    }
   }
   else {
-	$self->error("Found the JSON Data block and was "
-		. "able to parse it, but it had no location markers "
-		. "in it.  Maybe Google changed their "
-		. "JSON data structure?.");
-	return undef;
+    $self->error("Google couldn't resolve the address $address but suggested alternate addresses.  "
+      . "I attempted to download them but failed.") and return undef;
   }
+  return @result;
 }
 
 =head2 near()
@@ -300,35 +308,30 @@ sub near {
 
   # See if google returned no results
   if ( $page =~ /did\snot\smatch\sany\slocations/i ) {
-    $self->error( "Google couldn't find a $query near " . $where->title);
-    return undef;
+    $self->error( "Google couldn't find a $query near " . $where->title) and return undef;
+  }
+  # attempt to locate the JSON formatted data block
+  elsif ($page =~ m#loadVPage\((.+), "\w+"\);}//]]>#is) {
+    my $strJSON = $1;
+    $response_json = $json->jsonToObj($strJSON);
   }
-  # attept to locate the JSON formatted data block
-  elsif ($page =~ 	
-	/loadVPage\((.+),document\.getElementById\(/is) {
-	my $strJSON = $1;
-	$response_json = $json->jsonToObj($strJSON);
-	}
   else {
-	$self->error( "Unable to locate the JSON format data in " 
-		. "google's response.");
-	return undef;
-	}
+    $self->error( "Unable to locate the JSON format data in Google's response.") and return undef;
+  }
 
   if ( scalar(@{$response_json->{"overlays"}->{"markers"}}) > 0 ) {
-	my @result = ();
-	foreach my $marker (@{$response_json->{"overlays"}->{"markers"}}) {
-		my $loc = $self->_obj2location($marker);
-		push @result, $loc;
-	}		
-  	return @result;
+    my @result = ();
+    foreach my $marker (@{$response_json->{"overlays"}->{"markers"}}) {
+      my $loc = $self->_obj2location($marker);
+      push @result, $loc;
+    }		
+    return @result;
   }
   else {
-	$self->error("Found the JSON Data block and was "
-		. "able to parse it, but it had no location markers"
-		. "in it.  Maybe Google changed their "
-		. "JSON data structure?");
-	return undef;
+    $self->error("Found the JSON Data block and was "
+      . "able to parse it, but it had no location markers"
+      . "in it.  Maybe Google changed their "
+      . "JSON data structure?") and return undef;
   }
 }
 
@@ -369,31 +372,42 @@ sub path {
   for (my $i=1; $i<=$#locations; $i++){
 	$googlesearch .= " to:" . join(', ', $locations[$i]->lines);
   }
-
   my $page = get( sprintf( LQ, uri_escape( $googlesearch ) ) );
 
   # See if google returned no results
   if ( $page =~ /did\snot\smatch\sany\slocations/i ) {
-    $self->error( "Google couldn't find one of the locations"
-		. " you provided for your directions query");
-    return undef;
+    $self->error( "Google couldn't find one of the locations you provided for your directions query") and return undef;
+  }
+  # See if google didn't recognize an input, but suggested
+  # a correction to the input that it does recognize
+  elsif ( $page =~ m#didyou#s )
+  {
+    # Parse the JSON to unescape the escaped unicode characters in the URLs we need to parse
+    my ( $strJSON ) = $page =~ m#loadVPage\((.+), "\w+"\);}//]]>#s;
+    my $suggestion_json = $json->jsonToObj($strJSON);
+    # Did you mean:</span><div class="ref"><a href="/maps?v=1&amp;ie=UTF8&amp;hl=en&amp;ct=clnk&amp;cd=1&amp;saddr=695+Charles+E+Young+Dr+S,+Los+Angeles,+Los+Angeles,+California+90024,+United+States&amp;daddr=10948+Weyburn+Ave,+Los+Angeles,+CA+90024+to:5006+W+Pico+Blvd,+Los+Angeles,+CA+90019&amp;f=d" onclick="return loadUrl(this.href)"><b><i>695 Charles E Young Dr S, Los Angeles, Los Angeles, California 90024, United States</i></b>
+    my ( $first_suggestion ) = $suggestion_json->{panel} =~ m#(saddr=.+?)" onclick#s;
+    # Get the directions using google's first suggestion
+    $page = get ( _html_unescape("http://maps.google.com/maps?output=js&$1") );
+
+    # warn the user using the error method, but don't return undef.
+    $self->error("Google suggested a different address for your query.  Using the google suggestion instead.");
   }
   # attept to locate the JSON formatted data block
-  elsif ($page =~ 	
-	/loadVPage\((.+),document\.getElementById\(/is) {
-	my $strJSON = $1;
-
-	# Extract the JSON data structure from the response.
-	$response_json = $json->jsonToObj($strJSON);
-	}
+  if ($page =~ m#loadVPage\((.+), "\w+"\);}//]]>#s) {
+    # Extract the JSON data structure from the response.
+    $response_json = $json->jsonToObj( $1 );
+  }
   else {
-	$self->error( "Unable to locate the JSON format data in " 
-		. "google's response.");
-	return undef;
+    $self->error( "Unable to locate the JSON format data in Google's response.") and return undef;
   }
 
-  my $enc_points = $response_json->{"overlays"}->{"polylines"}->[0]->{"points"};
-  my @points = _decode($enc_points);
+  my @points;
+  my @enc_points;
+  for (my $i = 0; $i<=$#{$response_json->{"overlays"}->{"polylines"}}; $i++) {
+    $enc_points[$i] = $response_json->{"overlays"}->{"polylines"}->[$i]->{"points"};
+    $points[$i] = [ _decode($enc_points[$i]) ];
+  }
 
   # extract a series of directions from HTML inside the panel 
   # portion of the JSON data response, stuffing them in @html_segs
@@ -403,8 +417,8 @@ sub path {
   my $panel = $response_json->{'panel'};
   $panel =~ s/&#160;/ /g;
 
-  my @subpaths = $panel =~ m#(<table class=042(ddrsteps pw|ddwpt_table)042.+?</table>\s*</div>)#gs; #ddspt_table
-  #my ( $subpanel ) = $response_json->{'panel'} =~ m#<table class=042ddrsteps pw042>(.+)</table>#s;
+  my @subpaths = $panel =~ m#(<table class="(ddrsteps(?: pw)?|ddwpt_table|dirsegment)".+?</table>\s*</div>)#gs; #ddspt_table
+  #my ( $subpanel ) = $response_json->{'panel'} =~ m#<table class="ddrsteps pw">(.+)</table>#s;
 
   foreach my $subpath ( @subpaths ) {
     my @segments = split m#</tr>\s*<tr#s, $subpath;
@@ -412,25 +426,21 @@ sub path {
       #skip irrelevant waypoint rows
       if ( $subpath =~ m#ddwpt_table#s && $segment !~ m#ddptlnk#s ) { next }
 
-      my ( $id )         = $segment =~ m#id=042(.+?)042#s;
-      my ( $pointIndex ) = $segment =~ m#polypoint=042(.+?)042#s;
-      my ( $html )       = $segment =~ m#042dirsegtext042>(.+?)</td>#s;
-      my ( $distance )   = $segment =~ m#042sdist042>.*?<b>(.+?)</b>.*?</td>#s;
-      my ( $time )       = $segment =~ m#042segtime pw042>(.+?)<#s;
+      my ( $id, $pointIndex ) = $segment =~ m#id="(.+?)" polypoint="(.+?)"#s;
+      my ( $html )       = $segment =~ m#"dirsegtext_\d+_\d+">(.+?)</td>#s;
+      my ( $distance )   = $segment =~ m#"sxdist".+?>(.+?)<#s;
+      my ( $time )       = $segment =~ m#"segtime nw pw">(.+?)<#s;
 
       if ( ! defined( $id ) ) {
-        if ( $subpath =~ m#waypoint=042(.+?)042#s ) {
+        if ( $subpath =~ m#waypoint="(.+?)"#s ) {
           $id = "waypoint_$1";
+	  $html = $locations[$1]->title();
+          ($pointIndex)  = $segment =~ m#polypoint="(.+?)"#s;
         }
       }
 
       next unless $id;
 
-      if ( ! $html ) {
-        #destinations/waypoints are different
-        ( $html ) = $segment =~ m#042milestone042.+?>(.+?)<td class=042closer#s;
-      }
-
       if ( ! $time ) {
         #some segments are different (why? what is the pattern?)
         my ( $d2, $t2 ) = $segment =~ m#timedist ul.+?>(.+?)\(about&\#160;(.+?)\)</td>#s;
@@ -444,16 +454,6 @@ sub path {
       $distance   ||= ''; $distance = decode_entities( $distance ); $distance =~ s/\s+/ /g;
       $time       ||= ''; $time     = decode_entities( $time     ); $time =~ s/\s+/ /g;
 
-#print
-#"$id
-#	$pointIndex
-#	$distance
-#	$time
-#	$html
-#
-#";
-#warn
-
       push (@html_segs, {
         distance   => $distance,
         time       => $time,
@@ -464,34 +464,67 @@ sub path {
       $stepsfound++;
     }
   }
-#die;
 
   if ($stepsfound == 0) {
-	$self->error("Found the HTML directions from the JSON "
-			. "reponse, but was not able to extract "
-			. "the driving directions from the HTML");
-	return undef;
+    $self->error("Found the HTML directions from the JSON "
+      . "reponse, but was not able to extract "
+      . "the driving directions from the HTML") and return undef;
   }
   my @segments = ();
-  # The first point in the first segment should be the start point, 
-  # but google doesn't include that in their polyline, so I compensate
-  my @points_subset = ( $locations[0] );
-  my $m = -1;
-  #  Correlate the array of lats and longs we decoded from the 
+  # Problem:  When you create a Geo::Google::Location by
+  # looking it up on Google from an address, it returns coordinates
+  # with millionth of a degree precision.  Coordinates that come out 
+  # the polyline string only have hundred thousandth of a degree
+  # precision.  This means that the correlation algorithm won't find
+  # the start, stop or waypoints in the polyline unless we round
+  # start, stop and waypoint coordinates to the hundred-thousandth
+  # degree precision.
+  foreach my $location (@locations) {
+    $location->{'latitude'} = sprintf("%3.5f", $location->{'latitude'} );
+    $location->{'longitude'} = sprintf("%3.5f", $location->{'longitude'} );
+  }
+
+  #  Correlate the arrays of lats and longs we decoded from the 
   # JSON object with the segments we extracted from the panel 
   # HTML and put the result into an array of
-  # Geo::Google::Location objects 
-  while ( @points ) {
-      my $lat = shift @points;
-      my $lon = shift @points;
-      $m++;
+  # Geo::Google::Location objects
+  my @points_subset = ( $locations[0] );
+  push (@segments, Geo::Google::Segment->new(
+        pointIndex => $html_segs[0]{'pointIndex'},
+        id         => $html_segs[0]{'id'},
+        html       => $html_segs[0]{"html"},
+        distance   => $html_segs[0]{'distance'},
+        time       => $html_segs[0]{'time'},
+        from       => $locations[0],
+        to         => $locations[0],
+        points     => [@points_subset])
+	);
+  shift @html_segs;
+  for (my $i = 0; $i <= $#points; $i++) {
+    # start/points cause us problems because they're often the same
+    # the same pointindex as the first segment of the directions
+    # pulling the first html_seg off the stack now makes the next
+    # control loop easier to maintain.
+    @points_subset = ();
 
+    my $m = 0;
+    my @pointset = @{$points[$i]};
+    while ( @pointset ) {
+      my $lat = shift @pointset;
+      my $lon = shift @pointset;
+      $m++;
       my %html_seg;
 
-      my $point = Geo::Google::Location->new(
-          latitude  => $lat,
-          longitude => $lon,
-        );
+      # Check to see if the lat and long belong to a start, stop or waypoint
+      my $pointislocation = -1;
+      for (my $j=0; $j <= $#locations; $j++) {
+	if ( ( $lat == $locations[$j]->latitude() ) && ( $lon == $locations[$j]->longitude() ) ) { $pointislocation = $j; last; }
+      }
+      # If the point that just came off the pointset array is a start, stop or waypoint, use that start/stop/waypoint.
+      # otherwise, create a new point for the lat/long that just came off the pointset array.
+      my $point;
+      if ( $pointislocation >= 0 ){ $point = $locations[$pointislocation]; }
+      else			  { $point = Geo::Google::Location->new( latitude  => $lat, longitude => $lon ); }
 
       push @points_subset, $point;
 
@@ -499,11 +532,9 @@ sub path {
 	# There's a segment after the one we're working on
 	# This tests to see if we need to wrap up the current segment
         if ( defined( $html_segs[1]{'pointIndex'} ) ) {
-          next unless $m + 1 == $html_segs[1]{'pointIndex'};
+          next unless ((($m == $html_segs[1]{'pointIndex'}) && ($#html_segs > 1) ) || (! @pointset) );
         }
-
         %html_seg = %{shift @html_segs};
-
         push @segments, Geo::Google::Segment->new(
           pointIndex => $html_seg{'pointIndex'},
           id         => $html_seg{'id'},
@@ -517,8 +548,12 @@ sub path {
         @points_subset = ();
       } elsif ($html_segs[0]) { # We're working on the last segment
 	# This tests to see if we need to wrap up the last segment
-         next unless (! $points[0]);
+         next unless (! $pointset[0]);
          %html_seg = %{shift @html_segs};
+
+	 # An attempt to get the last point in the last segment
+	 # set.  Google doesn't include it in their polylines.
+	 push @points_subset, $locations[$i+1];
          push @segments, Geo::Google::Segment->new(
             pointIndex => $html_seg{'pointIndex'},
             id         => $html_seg{'id'},
@@ -526,19 +561,26 @@ sub path {
             distance   => $html_seg{'distance'},
             time       => $html_seg{'time'},
             from       => $points_subset[0],
-            to         => $locations[$#locations],
+            to         => $locations[$i+1],
             points     => [@points_subset]
           );
           @points_subset = ();
       } else { # we accidentally closed out the last segment early
           push @{ $segments[$#segments]->{points} }, $point;
       }
+    }
   }
-  # The last point in the last segment should be the final 
-  # destination, but google doesn't put that in the polyline.
-  push @{ $segments[$#segments]->{points} }, $locations[$#locations];
-    
-
+  # Dirty:  add the final waypoint
+  push (@segments, Geo::Google::Segment->new(
+          pointIndex => $html_segs[0]{'pointIndex'},
+          id         => $html_segs[0]{'id'},
+          html       => $html_segs[0]{"html"},
+          distance   => $html_segs[0]{'distance'},
+          time       => $html_segs[0]{'time'},
+          from       => $locations[$#locations],
+          to         => $locations[$#locations],
+          points     => [ ($locations[$#locations]) ])
+	);
   # Extract the total information using a regex on the panel hash
   # At the end of the "printheader", we're looking for:
   # 282&#160;mi&#160;(about&#160;4 hours 27 mins)</td></tr></table>
@@ -548,12 +590,11 @@ sub path {
 	   segments  => \@segments,
 	   distance  => $1 . " " . $2,
 	   time      => $3,
-	   polyline  => $enc_points,
+	   polyline  => [ @enc_points ],
 	   locations => [ @locations ],
 	   panel     => $response_json->{"panel"},
 	   levels    => $response_json->{"overlays"}->{"polylines"}->[0]->{"levels"}
 					);
-
 	return $path;
   } else {
 	$self->error("Could not extract the total route distance "
@@ -574,80 +615,91 @@ sub path {
 
 =cut
 
+=head2 _decode_word()
+
+ Usage    : my $float = _decode_word($encoded_quintet_word);
+ Function : turn a quintet word into a float for the _decode() function
+ Returns  : a float
+ Args     : one data word made of ASCII characters carrying
+            a five-bit number per character from an encoded 
+	    Google polyline string
+
+=cut
+
+sub _decode_word {
+  my $quintets = shift;
+  my @quintets = split '', $quintets;
+  my $num_chars = scalar(@quintets);
+  my $i = 0;
+  my $final_number = 0;
+  my $ordinal_offset = 63;
+  
+  while ($i < $num_chars ) {
+    if ( ord($quintets[$i]) < 95 ) { $ordinal_offset = 63; }
+    else { 		             $ordinal_offset = 95; }
+    my $quintet = ord( $quintets[$i] ) - $ordinal_offset;
+    $final_number |= $quintet << ( $i * 5 );
+    $i++;
+  }
+  if ($final_number % 2 > 0) { $final_number *= -1; $final_number --; }
+  return $final_number / 2E5;
+}
+
 =head2 _decode()
 
  Usage    : my @points = _decode($encoded_points);
  Function : decode a polyline into its composite lat/lon pairs
- Returns  : an array
- Args     : a string
+ Returns  : an array of floats (lat1, long1, lat2, long2 ... )
+ Args     : an encoded google polyline string
 
 =cut
 
 sub _decode {
-  my $points = shift;
-  return undef unless defined $points;
-  my @points = split '', $points;
-  my $Ch = scalar(@points);
-  my $pb = 0;
+  # Each letter in the polyline is a quintet (five bits in a row).
+  # A grouping of quintets that makes up a number we'll use
+  # to calculate lat and long will be called a "word".
+  my $quintets = shift;
+  return undef unless defined $quintets;
+  my @quintets = split '', $quintets;
   my @locations = ();
-  my $Ka = 0;
-  my $Pa = 0;
-  my $i = undef;
-  while ( $pb < $Ch ) {
-    my $oc = 0;
-    my $Fa = 0;
+  my $word = "";
 
-    while ( 1 ) {
-      my $ub = ord($points[$pb]) - 63;
-      $pb++;
-      $Fa |= ($ub & 31) << $oc;
-      $oc += 5;
-      last if $ub < 32;
+  # Extract the first lat and long.
+  # The initial latitude word is the first five quintets.
+  for (my $i=0; $i<=4; $i++) { $word .= $quintets[$i]; }
+  push ( @locations, _decode_word($word) );
+  my $lastlat = 0;
+
+  # The initial longitude is the next five quintets.
+  $word = "";
+  for (my $i=5; $i<10; $i++) { $word .= $quintets[$i]; }
+  push ( @locations, _decode_word($word) );
+  my $lastlong = 1;
+
+  # The remaining quintets form words that represent 
+  # delta coordinates from the last coordinate.  The only
+  # way to identify them is that they are at least one
+  # character long and end in a ASCII character between
+  # ordinal 63 and ordinal 95.  Latitude first, then
+  # longitude.
+  $word = "";
+  my $i = 10;
+  while ($i <= $#quintets) {
+    $word .= $quintets[$i];
+    if ( (length($word) >= 1) && ( ord($quintets[$i]) <= 95 ) ) {
+      if ( $lastlat > $lastlong ) {
+        push @locations, _decode_word($word) + $locations[$lastlong];
+	$lastlong = $#locations; 
+      }
+      else {
+        push @locations, _decode_word($word) + $locations[$lastlat];
+	$lastlat = $#locations; 
+      }
+      $word = "";
     }
-
-    if ( $Fa & 1 ) {
-      $i = ~($Fa >> 1);
-    } else {
-      $i = $Fa >> 1;
-    }
-
-    $Ka += $i;
-    push @locations, ($Ka * 1E-5);
-
-    #negative values come out wrong -- some bitshift error, will fix later.
-    #it's a hack, but it works.
-    while ( $locations[-1] >= 42000 ) {
-      $locations[-1] -= 42949.67296;
-    }
-
-    $oc = 0;
-    $Fa = 0;
-
-    while ( 1 ) {
-      my $ub = ord($points[$pb]) - 63;
-      $pb++;
-      $Fa |= ($ub & 31) << $oc;
-      $oc += 5;
-      last if $ub < 32;
-    }
-    if ( $Fa & 1 ) {
-      $i = ~($Fa >> 1);
-    } else {
-      $i = $Fa >> 1;
-    }
-
-    $Pa += $i;
-    push @locations, ($Pa * 1E-5);
-
-    #negative values come out wrong -- some bitshift error, will fix later.
-    #it's a hack, but it works.
-    while ( $locations[-1] >= 42000 ) {
-      $locations[-1] -= 42949.67296;
-    }
-
+    $i++;
   }
-
-  #prettify results
+  # Prettify results
   return map {sprintf("%3.5f",$_)} @locations;
 }
 
@@ -656,51 +708,63 @@ sub _decode {
  Usage    : my $encoded_points = _encode(@points);
  Function : encode lat/lon pairs into a polyline string
  Returns  : a string
- Args     : an array
+ Args     : an array of coordinates [38.47823, -118.48571, 38.47845, -118.48582, ...]
 
 =cut
 
 sub _encode {
   my @points = @_;
-  my $numPoints = scalar(@points);
-
-  @points = map { int($_/0.00001) } @points;
-  my @l = ();
-  my $xo = 0;
-  my $yo = 0;
-
-  foreach my $i ( 0..$numPoints-1 ) { 
-    my $y = $points[ $i << 1 ];
-    my $yd = $y - $yo;
-    $yo = $y;
-
-    my $f = ( abs($yd) << 1 ) - ( $yd < 0 );
-
-    while ( 1 ) {
-      my $e = $f & 31;
-      $f >>= 5;
-      if ( $f ) {
-        $e |= 32;
-      }
-      push @l, chr($e+63);
-      last if $f == 0;
+  my $polyline;
+  for (my $i = 0; $i <= $#points; $i++) {
+    # potential pitfall: pass the correct floating point precision
+    # to the _encode_word() function or 34.06694 - 34.06698 will give you
+    # -3.999999999999999057E-5 which doesn't encode properly. -4E-5 encodes properly.
+    if ( $i > 1 ) { # All points after the first lat/long pair are delta coordinates
+      $polyline .= _encode_word( sprintf("%3.5f", $points[$i] - $points[$i-2] ) );
     }
-
-    my $x = $points[ ($i << 1) + 1 ];
-    my $xd = $x - $xo;
-    $xo = $x;
-    $f = ( abs($xd) << 1 ) - ( $xd < 0 );
-    while ( 1 ) {
-      my $e = $f & 31;
-      $f >>= 5;
-      if ( $f ) { 
-        $e |= 32;
-      }
-      push @l, chr($e+63);
-      last if $f == 0;
+    else {
+      $polyline .= _encode_word( sprintf("%3.5f", $points[$i] ) );
     }
   }
-  return join '', @l;
+  return $polyline;
+}
+
+=head2 _encode_word()
+
+ Usage    : my $encoded_quintet_word = _encode_word($signed_floating_point_coordinate);
+ Function : turn a signed float (either a full coordinate 
+	    or a delta) for the _encode() function
+ Returns  : a string containing one encoded coordinate that
+	    will be added to a polyline string
+ Args     : one data word made of ASCII characters carrying
+            a five-bit number per character from an encoded 
+	    Google polyline string
+
+=cut
+
+sub _encode_word {
+  my $coordinate = shift;
+  # Convert the floating point coordinate into a doubled signed integer.  -38.45671 turns into -7691342
+  # This looks quirky cos when I used int(-0.00015 * 2E5) I got -29 (should have been -30).  Suspect this is a perl 5.8.8 bug (MAT).
+  my $signed_int = int( sprintf("%8.0f", $coordinate * 2E5) );
+  # If the signed integer is negative, add one then lose the sign.  -7691342 turns into 7691341
+  my $unsigned_int;
+  if ($signed_int < 0) { $unsigned_int = -($signed_int + 1); }
+  else		       { $unsigned_int = $signed_int;	     }
+  
+  # Quintets get created in reverse order (least signficant quintet first, most significant quintet last)
+  my $ordinal_offset;
+  my $quintet;
+  
+  # This do...while structure allows me to properly encode the coordinate 0
+  do {
+    if ( $unsigned_int < 32 ) { $ordinal_offset = 63; } #last quintet
+    else { 		        $ordinal_offset = 95; }
+    my $quintet_mask = ( $unsigned_int >> 5 ) << 5;
+    $quintet .= chr( ( $unsigned_int ^ $quintet_mask ) + $ordinal_offset );
+    $unsigned_int = $unsigned_int >> 5;
+  } while ( $unsigned_int > 0 );
+  return $quintet;
 }
 
 =head2 _html_unescape()
@@ -745,23 +809,22 @@ sub _obj2location {
   # Check to make sure that the info window contents are HTML
   # and that google hasn't changed the format since I wrote this
   if ( $marker->{"infoWindow"}->{"type"} eq "html" ) {
-	if ($marker->{"laddr"} =~ 
-		/\((.+)\)\s\@\-?\d+\.\d+,\-?\d+\.\d+$/s){
-		$title = $1;
-	}
-	else {
-		$title = $marker->{"laddr"};
-	}
+    if ($marker->{"laddr"} =~ /\((.+)\)\s\@\-?\d+\.\d+,\-?\d+\.\d+$/s){
+      $title = $1;
+    }
+    else {
+      $title = $marker->{"laddr"};
+    }
 
-	$description = decode_entities($marker->{"infoWindow"}->{"basics"});
-	# replace </P>, <BR>, <BR/> and <BR /> with newlines
-	$description =~ s/<\/p>|<br\s?\/?>/\n/gi;
-	# remove all remaining markup tags
-	$description =~ s/<.+>//g;
+    $description = decode_entities($marker->{"infoWindow"}->{"basics"});
+    # replace </P>, <BR>, <BR/> and <BR /> with newlines
+    $description =~ s/<\/p>|<br\s?\/?>/\n/gi;
+    # remove all remaining markup tags
+    $description =~ s/<.+>//g;
   }
   else {
-	# this is a non-fatal nuisance error, only lat/long are 
-	# absolutely essential products of this function
+    # this is a non-fatal nuisance error, only lat/long are 
+    # absolutely essential products of this function
     $title = "Could not extract a title or description from "
 	. "google's response.  Have they changed their format since "
 	. "this function was written?";
@@ -771,7 +834,7 @@ sub _obj2location {
     title     => $title,
     latitude  => $marker->{"lat"},
     longitude => $marker->{"lng"},
-    lines     => [split(/\n/, $description)],
+    lines     => [ @{ $marker->{"addressLines"} } ],
     id        => $marker->{"id"}
                  || $arg{'id'}
                  || md5_hex( localtime() ),
